#include <iostream>
#include <cmath>

using namespace std;

#include "expr.hh"
#include "basfun.hh"
#include "stmt.hh"
#include "typenameof.hh"

expr::expr()
{
}
expr::~expr()
{
}

static const std::string TypeToString(TypeP p)
{
    return p->Dump();
}

tristate expr::is_num() const
{
    TypeP t = GetType();
#ifdef BINDING_DEBUG
    //dprintf(dp_declare, "is_num? %s\n", t?TypeToString(t).c_str():"NULL");
#endif
    if(!t) return maybe; // unknown
    return !t->is_array() && t->is_num();
}

tristate expr::is_string() const
{
    TypeP t = GetType();
#ifdef BINDING_DEBUG
    //dprintf(dp_declare, "is_string? %s\n", t?TypeToString(t).c_str():"NULL");
#endif
    if(!t) return maybe; // unknown
    return !t->is_array() && t->is_string();
}

bool expr_number::is_int() const
{
    if(!type) return false;
    if(type->is_array()) return false;
    return type->Type == TypeInfo::Integer || type->Type == TypeInfo::Long
        || float_val == int_val;
}
long double   expr_number::get_f() const { return is_int() ? int_val : float_val; }
long long int expr_number::get_i() const { return is_int() ? int_val : BasCINT(float_val); }


TypeP expr_unary::GetType() const 
{
    TypeP t = p->GetType();
    if(t && token == NOT)
        return t->Type == TypeInfo::Long  ?  t  :  TypeInt;
    return t;
}

exprp expr_unary::optimized()
{
    expr_number* n = dynamic_cast<expr_number*> (&*p);
    switch(token)
    {
        case '-': if(n) return new expr_number(- n->get_i()); break;
        case NOT: if(n) return new expr_number(~ n->get_i()); break;
        default: break;
    }
    return expr::optimized();
}

TypeP expr_binary::GetType() const
{
    // choices:
    //   OR, XOR, AND, comparisons, plus, minus, MOD, '\\', mul, div, '^'
    
    tristate result_integer = false;
    switch(token)
    {
        case OR: case XOR: case AND: case '>': case '<': case '=':
        case NOTEQUAL: case LESSOREQUAL: case GREATEROREQUAL:
        case MOD: case '\\':
            result_integer = true; break;
        case '+': case '-': case '*':
            result_integer = maybe; break;
        default: break;
    }
    TypeP a = p0->GetType(); if(!a) return a;
    TypeP b = p1->GetType(); if(!b) return b;
    
    if(result_integer.is_true())
    {
        if(a->Type == TypeInfo::Long) return a;
        if(b->Type == TypeInfo::Long) return b;
        return TypeInt;
    }
    if(result_integer.is_maybe())
    {
        int a_int_level =
           (a->Type == TypeInfo::Long ? 2 : a->Type == TypeInfo::Integer ? 1 : 0);
        int b_int_level =
           (b->Type == TypeInfo::Long ? 2 : b->Type == TypeInfo::Integer ? 1 : 0);
        if(a_int_level && b_int_level)
            return max(a_int_level,b_int_level)==2 ? TypeLng : TypeInt;
    }
    
    if(a->Type == TypeInfo::String || b->Type == TypeInfo::String) return TypeStr;
    if(a->Type == TypeInfo::Single && b->Type == TypeInfo::Single) return TypeSng;
    if(a->Type == TypeInfo::UserType || b->Type == TypeInfo::UserType) return 0;
    return TypeDbl;
}

bool expr_binary::IsComparison() const
{
    switch(token)
    {
        case '<':
        case '>':
        case '=':
        case LESSOREQUAL:
        case GREATEROREQUAL:
        case NOTEQUAL:
            return true;
        case AND:
        case OR:
            return p0->IsComparison() || p1->IsComparison();
        default: break;
    }
    return false;
}

exprp expr_binary::optimized()
{
    expr_string* s0 = dynamic_cast<expr_string*> (&*p0);
    expr_string* s1 = dynamic_cast<expr_string*> (&*p1);
    switch(token)
    {
        #define stringop(t,op,p) if(s0&&s1) return new t((s0->get() op s1->get()) p); break
        case '+':            stringop(expr_string,+,);
        case '<':            stringop(expr_number,<,?-1:0);
        case '>':            stringop(expr_number,>,?-1:0);
        case '=':            stringop(expr_number,==,?-1:0);
        case LESSOREQUAL:    stringop(expr_number,<=,?-1:0);
        case GREATEROREQUAL: stringop(expr_number,>=,?-1:0);
        case NOTEQUAL:       stringop(expr_number,!=,?-1:0);
        #undef stringop
        default: break;
    }
    
    expr_number* n0 = dynamic_cast<expr_number*> (&*p0);
    expr_number* n1 = dynamic_cast<expr_number*> (&*p1);
    switch(token)
    {
        #define numberop(t,op,p) if(n0&&n1) return new t((n0->get_f() op n1->get_f()) p); break
        case '+': numberop(expr_number,+,);
        case '-': numberop(expr_number,-,);
        case '*': numberop(expr_number,*,);
        case '/': numberop(expr_number,/,);
        
        case '^': if(n0&&n1) return new expr_number(std::pow(n0->get_f(), n1->get_f())); break;

        case '<':            numberop(expr_number,<,?-1:0);
        case '>':            numberop(expr_number,>,?-1:0);
        case '=':            numberop(expr_number,==,?-1:0);
        case LESSOREQUAL:    numberop(expr_number,<=,?-1:0);
        case GREATEROREQUAL: numberop(expr_number,>=,?-1:0);
        case NOTEQUAL:       numberop(expr_number,!=,?-1:0);
        
        case MOD: if(n0&&n1) return new expr_number(BasMODop(n0->get_i(), n1->get_i()));
        
        case XOR: if(n0&&n1) return new expr_number(n0->get_i() ^ n1->get_i());
        case OR:  if(n0&&n1) return new expr_number(n0->get_i() | n1->get_i());
        case AND: if(n0&&n1) return new expr_number(n0->get_i() & n1->get_i());
        
        case '\\': if(n0&&n1) return new expr_number(BasINT(n0->get_i() / n1->get_i()));
        
        #undef numberop
        
        default: break;
    }
    return expr::optimized();
}

exprp expr_builtin_numeric::optimized()
{
    expr_number* n0 =
        params.size() >= 1
        ? dynamic_cast<expr_number*> (&*params[0])
        : 0;
    expr_string* s0 =
        params.size() >= 1
        ? dynamic_cast<expr_string*> (&*params[0])
        : 0;
    
    switch(token)
    {
        case ABS: if(n0) return new expr_number(fabsl(n0->get_f()));
        case ATN: if(n0) return new expr_number(atanl(n0->get_f()));
        case CINT: if(n0) return new expr_number(n0->get_i(), TypeInt);
        case CLNG: if(n0) return new expr_number(n0->get_i(), TypeLng);
        case CSNG: if(n0) return new expr_number(n0->get_f(), TypeSng);
        case CDBL: if(n0) return new expr_number(n0->get_f(), TypeDbl);
        case COS: if(n0) return new expr_number(cosl(n0->get_f()));
        case EXP: if(n0) return new expr_number(expl(n0->get_f()));
        case FIX: if(n0) return new expr_number(truncl(n0->get_f()));
        case LOG: if(n0) return new expr_number(logl(n0->get_f()));
        case SQR: if(n0) return new expr_number(sqrtl(n0->get_f()));
        case TAN: if(n0) return new expr_number(tanl(n0->get_f()));
        case VAL: if(s0) return new expr_number(BasVAL(s0->get()));
        default: break;
    }
    return expr::optimized();
}

TypeP expr_builtin_numeric::GetType() const
{
    switch(token)
    {
        case CINT: return TypeInt;
        case CLNG: return TypeLng;
        case CSNG: return TypeSng;
        case CDBL: return TypeDbl;
        default: break;
    }
    return TypeDbl;
}

exprp expr_builtin_string::optimized()
{
    expr_number* n0 =
        params.size() >= 1
        ? dynamic_cast<expr_number*> (&*params[0])
        : 0;

    expr_number* n1 =
        params.size() >= 2
        ? dynamic_cast<expr_number*> (&*params[1])
        : 0;

    expr_string* s0 =
        params.size() >= 1
        ? dynamic_cast<expr_string*> (&*params[0])
        : 0;

#if 0

        /* Pure functions: functions that only depend on their parameter;
         * can be calculated at compile-time. */
        return (token == LTRIM_S
             || token == RTRIM_S
             || token == LCASE_S
             || token == UCASE_S
             || token == INSTR
             || token == MID_S
             || token == RIGHT_S
             || token == STRING_S
             || token == CHR_S
             || token == HEX_S
             || token == MKD_S
             || token == MKI_S
             || token == MKL_S
             || token == MKS_S
             || token == OCT_S
             || token == SPACE_S
             || token == STR_S
               );
#endif
    switch(token)
    {
        case LTRIM_S: if(s0) return new expr_string(BasLTRIM(s0->get()));
        case RTRIM_S: if(s0) return new expr_string(BasRTRIM(s0->get()));
        case LCASE_S: if(s0) return new expr_string(BasLCASE(s0->get()));
        case UCASE_S: if(s0) return new expr_string(BasUCASE(s0->get()));
        
        case STR_S: if(n0) return new expr_string(BasSTRs(n0->get_f()));
        case CHR_S: if(n0) return new expr_string(BasCHRs(n0->get_i()));
        
        default: break;
    }
    return expr::optimized();
}

autoptr<expr_userfun> expr_var::convert_to_functioncall(TypeP type) const
{
    return new expr_userfun(name, type, indexes);
}


void expr::Compile(ExprCompileDefs)
{
    std::cout << "    /* Compilation of "
              << typenameof(*this)
              << " not supported */\n";
}

void expr_string::ForEachSubExpr(expr_visitor& func)
{
    expr::ForEachSubExpr(func);
}

void expr_number::ForEachSubExpr(expr_visitor& func)
{
    expr::ForEachSubExpr(func);
}

void expr_unary::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(p);
    expr::ForEachSubExpr(func);
}

void expr_binary::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(p0);
    func.expr_visit(p1);
    expr::ForEachSubExpr(func);
}

void expr_label::ForEachSubExpr(expr_visitor& func)
{
    expr::ForEachSubExpr(func);
}

void expr_userfun::ForEachSubExpr(expr_visitor& func)
{
    for(unsigned a=0; a<params.size(); ++a)
    {
        func.set_func_call(name, a);
        func.expr_visit(params[a], maybe);
    }
    expr::ForEachSubExpr(func);
}

void expr_builtin_numeric::ForEachSubExpr(expr_visitor& func)
{
    for(unsigned a=0; a<params.size(); ++a)
        func.expr_visit(params[a],
          lvaluemap.size() > a && lvaluemap[a]);
    expr::ForEachSubExpr(func);
}

void expr_builtin_string::ForEachSubExpr(expr_visitor& func)
{
    for(unsigned a=0; a<params.size(); ++a)
        func.expr_visit(params[a],
          lvaluemap.size() > a && lvaluemap[a]);
    expr::ForEachSubExpr(func);
}

void expr_caseelse::ForEachSubExpr(expr_visitor& func)
{
    expr::ForEachSubExpr(func);
}

void expr_caseis::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(ref);
    expr::ForEachSubExpr(func);
}

void expr_caserange::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(min);
    func.expr_visit(max);
    expr::ForEachSubExpr(func);
}

void expr_caseor::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(a);
    func.expr_visit(b);
    expr::ForEachSubExpr(func);
}

void expr_var::ForEachSubExpr(expr_visitor& func)
{
    for(unsigned a=0; a<indexes.size(); ++a)
        func.expr_visit(indexes[a]);
    // ignore member
    expr::ForEachSubExpr(func);
}

void expr_returnvalueholder::ForEachSubExpr(expr_visitor& func)
{
    expr_var::ForEachSubExpr(func);
}

void expr_lubound::ForEachSubExpr(expr_visitor& func)
{
    func.expr_visit(var);
    func.expr_visit(num_index);
    expr::ForEachSubExpr(func);
}
